#include "opencv2/opencv.hpp"

#include "../hist_hue_sat.hpp"
#include "../mouse_rect.hpp"

float matchHistogram(const cv::Mat& hsv_img, const HistHueSat& hist_model,
                    cv::RotatedRect& trect, int search_margin, cv::Mat& bcmap)
{
    bcmap = cv::Mat::zeros(hsv_img.rows, hsv_img.cols, CV_32F);
    int ww = (int)(trect.size.width / 2);
    int hh = (int)(trect.size.height / 2);
    int cx = (int)trect.center.x;
    int cy = (int)trect.center.y;
    int xmin = cv::max(cx - ww, 0);
    int xmax = cv::min(cx + ww, hsv_img.cols - 1);
    int ymin = cv::max(cy - hh, 0);
    int ymax = cv::min(cy + hh, hsv_img.rows - 1);

    HistHueSat hist_cand;
    cv::RotatedRect rect_cand;
    rect_cand.size.width = trect.size.width;
    rect_cand.size.height = trect.size.height;
    float bc_max = 0.0;

    for (int j = ymin; j < ymax; j++) {
        for (int i = xmin; i < xmax; i++) {
            rect_cand.center.x = (float)i;
            rect_cand.center.y = (float)j;
            hist_cand.calc(hsv_img, rect_cand);
            float bc = hist_model.coeffBhattacharyya(hist_cand);
            bcmap.at<float>(j, i) = bc;
            if (bc > bc_max) {
                bc_max = bc;
                trect = rect_cand;
            }
        }
    }

    return bc_max;
}

int main() {
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("histogram", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("similarity map", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input, input_hsv, output, bcmap;
    cv::Mat disp_hist = cv::Mat(256, 256 * 2 + 1, CV_8UC3);
    disp_hist = cv::Scalar(128, 128, 128);

    HistHueSat hist_model;
    hist_model.visualizeLegend(disp_hist(cv::Rect(0, 0, 256, 256)));
    cv::imshow("histogram", disp_hist);

    enum TrackState tstate = TRACK_OFF;
    cv::RotatedRect trect;
    struct MouseParam mparam = { &trect, &tstate };
    cv::setMouseCallback("output", onMouse, &mparam);

    while (1) {
        cap >> input;
        cv::cvtColor(input, input_hsv, CV_BGR2HSV);
        input.copyTo(output);

        if (tstate == TRACK_INIT) {
            hist_model.calc(input_hsv, trect);
            tstate = TRACK_ON;

            hist_model.visualize(disp_hist(cv::Rect(256 + 1, 0, 256, 256)));
            cv::imshow("histogram", disp_hist);

        } else if (tstate == TRACK_SELECT
                   && trect.size.width > 0 && trect.size.height > 0) {
            cv::ellipse(output, trect, CV_RGB(255, 0, 0), 3);

        } else if (tstate == TRACK_ON) {
            float b = matchHistogram(input_hsv, hist_model, trect, 32, bcmap);
            if (b > 0.5) {
                cv::ellipse(output, trect, CV_RGB(255, 0, 0), 3);
            } else {
                cv::ellipse(output, trect, CV_RGB(0, 255, 0), 1);
            }
            cv::imshow("similarity map", bcmap);
        }

        cv::imshow("output", output);
        int key;
        if ((key = cv::waitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;  
            }
        }
    }

    return 0;
}
